@@ -314,6 +314,33 @@ describe("Validation", () => {
     },
   });
 
+  createSuccessTestCase("enum with undefinedAsNull", {
+    // eslint-disable-next-line no-undefined
+    enumKeywordAndUndefined: undefined,
+  });
+
+  createSuccessTestCase("enum with undefinedAsNull #2", {
+    enumKeywordAndUndefined: 0,
+  });
+
+  createSuccessTestCase("array with enum and undefinedAsNull", {
+    arrayStringAndEnum: ["a", "b", "c"],
+  });
+
+  createSuccessTestCase("array with enum and undefinedAsNull #2", {
+    // eslint-disable-next-line no-undefined
+    arrayStringAndEnum: [undefined, false, undefined, 0, "test", undefined],
+  });
+
+  createSuccessTestCase("array with enum and undefinedAsNull #3", {
+    // eslint-disable-next-line no-undefined
+    arrayStringAndEnum: [undefined, null, false, 0, ""],
+  });
+
+  createSuccessTestCase("string and undefinedAsNull #3", {
+    stringTypeAndUndefinedAsNull: "test",
+  });
+
   // The "name" option
   createFailedTestCase(
     "webpack name",
@@ -462,7 +489,7 @@ describe("Validation", () => {
           error.children.some(
             (child) =>
               child.keyword === "absolutePath" &&
-              child.dataPath === ".output.filename"
+              child.instancePath === "/output/filename"
           )
         ) {
           return (
@@ -2971,4 +2998,62 @@ describe("Validation", () => {
     {},
     webpackSchema
   );
+
+  createFailedTestCase(
+    "formatExclusiveMaximum #1",
+    {
+      formatExclusiveMaximum: "2016-02-05",
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
+
+  createFailedTestCase(
+    "formatExclusiveMaximum #2",
+    {
+      formatExclusiveMaximum: "2016-12-27",
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
+
+  createFailedTestCase(
+    "enum and undefinedAsNull",
+    {
+      enumKeywordAndUndefined: "foo",
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
+
+  createFailedTestCase(
+    "array with enum and undefinedAsNull",
+    {
+      arrayStringAndEnum: ["foo", "bar", 1],
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
+
+  createFailedTestCase(
+    "array with enum and undefinedAsNull #2",
+    {
+      // eslint-disable-next-line no-undefined
+      arrayStringAndEnum: ["foo", "bar", undefined, 1],
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
+
+  createFailedTestCase(
+    "array with enum and undefinedAsNull #3",
+    {
+      // eslint-disable-next-line no-undefined
+      arrayStringAndEnumAndNoUndefined: ["foo", "bar", undefined],
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
+
+  createFailedTestCase(
+    "string and undefinedAsNull",
+    {
+      stringTypeAndUndefinedAsNull: 1,
+    },
+    (msg) => expect(msg).toMatchSnapshot()
+  );
 });
